#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
LeLESK utilities
"""

# This code is a part of lelesk library: https://github.com/letuananh/lelesk
# :copyright: (c) 2014 Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

from texttaglib.chirptext.cli import setup_logging
setup_logging('logging.json', 'logs')


from nltk.tokenize import word_tokenize
from nltk import pos_tag
from nltk.stem import WordNetLemmatizer

wnl = WordNetLemmatizer()


PUNCS = '''[][!"#$%&'()*+,./:;<=>?@\\^_`{|}~-]“”'''


def ptpos_to_wn(ptpos, default='x'):
    ''' Penn Treebank Project POS to WN '''
    # Ref: http://www.ling.upenn.edu/courses/Fall_2003/ling001/penn_treebank_pos.html
    # CC Coordinating conjunction
    # CD Cardinal number
    # DT Determiner
    # EX Existential there
    # FW Foreign word
    # IN Preposition or subordinating conjunction
    # -----------------------------
    # JJ Adjective
    # JJR Adjective, comparative
    # JJS Adjective, superlative
    # -----------------------------
    # LS List item marker
    # MD Modal
    # -----------------------------
    # NN Noun, singular or mass
    # NNS Noun, plural
    # NNP Proper noun, singular
    # NNPS Proper noun, plural
    # -----------------------------
    # PDT Predeterminer
    # POS Possessive ending
    # PRP Personal pronoun
    # PRP$ Possessive pronoun
    # -----------------------------
    # RB Adverb
    # RBR Adverb, comparative
    # RBS Adverb, superlative
    # -----------------------------
    # RP Particle
    # SYM Symbol
    # TO to
    # UH Interjection
    # -----------------------------
    # VB Verb, base form
    # VBD Verb, past tense
    # VBG Verb, gerund or present participle
    # VBN Verb, past participle
    # VBP Verb, non­3rd person singular present
    # VBZ Verb, 3rd person singular present
    # -----------------------------
    # WDT Wh­determiner
    # WP Wh­pronoun
    # WP$ Possessive wh­pronoun
    # WRB Wh­adverb
    if ptpos.startswith('JJ'):
        return 'a'
    elif ptpos.startswith('NN'):
        return 'n'
    elif ptpos.startswith('RB'):
        return 'r'
    elif ptpos.startswith('VB'):
        return 'v'
    else:
        return default


def tokenize(sent):
    words = word_tokenize(sent)
    tags = pos_tag(words)
    tokens = [(w, t, wnl.lemmatize(w, pos=ptpos_to_wn(t, default='n'))) for w, t in tags]
    return tokens



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
A tool for converting Gloss WordNet into SQLite
Latest version can be found at https://github.com/letuananh/lelesk

@author: Le Tuan Anh <tuananh.ke@gmail.com>
'''

# Copyright (c) 2014, Le Tuan Anh <tuananh.ke@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

__author__ = "Le Tuan Anh <tuananh.ke@gmail.com>"
__copyright__ = "Copyright 2014, lelesk"
__credits__ = []
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Le Tuan Anh"
__email__ = "<tuananh.ke@gmail.com>"
__status__ = "Prototype"

# -------------------------------------------------------------------

import os
import unittest
from lelesk import LeLeskWSD, LeskCache


TEST_DIR = os.path.dirname(__file__)
TEST_CACHE = os.path.join(TEST_DIR, 'data', 'test_cache.db')


class TestMain(unittest.TestCase):

    def dump_scores(self, scores):
        for candidate, score, freq in scores:
            print("Candidate: {c} | score: {s} | freq: {f}".format(c=candidate.synset, s=score, f=freq))

    def test_data(self):
        return 'fish', 'there are so many fish in the river'

    def test_basic(self):
        print("Test LeLesk WSD")
        # without using cache
        #
        l = LeLeskWSD()
        w, sent = self.test_data()
        scores = l.lelesk_wsd(w, sent, context=sent.split())
        self.dump_scores(scores)
        self.assertEqual(scores[0].candidate.synset.ID, '02512053-n')
        self.assertEqual(scores[0].score, 2)
        self.assertEqual(scores[0].freq, 12)
        # wit tokenizer
        #
        l = LeLeskWSD()
        w, sent = self.test_data()
        scores = l.lelesk_wsd(w, sent)
        self.dump_scores(scores)
        self.assertEqual(scores[0].candidate.synset.ID, '02512053-n')
        self.assertEqual(scores[0].score, 2)
        self.assertEqual(scores[0].freq, 12)

        # With using cache
        #
        l = LeLeskWSD(dbcache=LeskCache(TEST_CACHE))
        scores = l.lelesk_wsd(w, sent, context=sent.split())
        self.dump_scores(scores)
        self.assertEqual(scores[0].candidate.synset.ID, '02512053-n')
        self.assertEqual(scores[0].score, 2)
        self.assertEqual(scores[0].freq, 12)

    def test_cache(self):
        l = LeLeskWSD(dbcache=":memory:")
        l.dbcache.select('01775164-v')

    def test_mfs(self):
        print("Test MFS WSD")
        # without cache
        l = LeLeskWSD()
        l.connect()
        w, sent = self.test_data()
        scores = l.mfs_wsd(w, sent)
        self.assertEqual(scores[0].candidate.synset.ID, '02512053-n')
        self.assertEqual(scores[0].score, 12)
        self.assertEqual(scores[0].freq, 12)

        # with cache
        l = LeLeskWSD(dbcache=LeskCache(TEST_CACHE))
        l.connect()
        w, sent = self.test_data()
        scores = l.mfs_wsd(w, sent)
        self.assertEqual(scores[0].candidate.synset.ID, '02512053-n')
        self.assertEqual(scores[0].score, 12)
        self.assertEqual(scores[0].freq, 12)
        # self.dump_scores(scores)

    def test_fast_wsd(self):
        # disable lemmatizer => faster
        l = LeLeskWSD()
        l.connect()
        w, sent = self.test_data()
        scores = l.mfs_wsd(w, sent, lemmatizing=False)
        self.assertTrue(scores)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
Test utility functions
Latest version can be found at https://github.com/letuananh/lelesk

@author: Le Tuan Anh <tuananh.ke@gmail.com>
'''

# Copyright (c) 2014, Le Tuan Anh <tuananh.ke@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

__author__ = "Le Tuan Anh <tuananh.ke@gmail.com>"
__copyright__ = "Copyright 2014, lelesk"
__credits__ = []
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Le Tuan Anh"
__email__ = "<tuananh.ke@gmail.com>"
__status__ = "Prototype"

# -------------------------------------------------------------------

import os
import unittest
import logging
from lelesk import LeLeskWSD, LeskCache
from lelesk.util import tokenize


################################################################
# Configuration
################################################################
TEST_DIR = os.path.dirname(__file__)
TEST_CACHE = os.path.join(TEST_DIR, 'data', 'test_cache.db')


def get_logger():
    return logging.getLogger(__name__)


class TestMain(unittest.TestCase):

    def test_known_concepts(self):
        wsd = LeLeskWSD()
        ss = wsd.smart_synset_search('Ali Baba', pos=None, deep_select=True)
        self.assertTrue(ss)
        self.assertIn('Ali Baba', list(ss)[0].lemmas)

    def test_synset_search(self):
        wsd = LeLeskWSD()
        nsses = wsd.smart_synset_search('love', 'n')
        vsses = wsd.smart_synset_search('love', 'v')
        self.assertEqual(len(nsses), 6)  # There should be 6 noun synsets
        self.assertEqual(len(vsses), 4)
        doge = wsd.smart_synset_search('doge', 'n')
        self.assertEqual(len(doge), 1)
        # verb?
        speckled = wsd.smart_synset_search('speckle', 'v')
        self.assertEqual(len(speckled), 2)

    def test_lelesk_set(self):
        wsd = LeLeskWSD()
        fish_synsets = wsd.smart_synset_search('fish', 'n')
        n02512053 = fish_synsets['02512053-n']
        leset = wsd.build_lelesk_set(n02512053.ID)
        get_logger().warning(leset)
        self.assertEqual(len(leset), 192)
        self.assertIn('fish', leset)
        self.assertIn('aquatic', leset)
        self.assertIn('shark', leset)
        # now with cache
        wsd = LeLeskWSD(dbcache=LeskCache())
        lesetc = wsd.build_lelesk_set(n02512053.ID)
        self.assertEqual(set(leset), set(lesetc))
        # TODO: Test this method properly

    def test_tokenize(self):
        text = 'I have eaten some cakes and gave my dogs some too.'
        actual = tokenize(text)
        print(actual)

    def test_stopwords(self):
        wsd = LeLeskWSD()
        print(wsd.stopwords)

    def test_caching(self):
        wsd = LeLeskWSD()
        # make sure that we have a token list
        fish_synsets = wsd.smart_synset_search('fish', 'n')
        n02512053 = fish_synsets['02512053-n']
        leset = wsd.build_lelesk_set(n02512053.ID)
        self.assertEqual(len(leset), 192)

        # try to cache our token list now ...
        DB_PATH = os.path.abspath('./data/test_leskcache.db')
        print('Test DB loc: {}'.format(DB_PATH))
        l = LeskCache(db_file=DB_PATH)
        # l = LeskCache()
        l.cache(str(n02512053.ID), leset)
        l.cache(n02512053.ID, leset)  # cache this twice
        ls = l.select(n02512053.ID)
        self.assertEqual(len(ls), 192)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
Le's LESK - Word Sense Disambiguation Package
'''

# This code is a part of lelesk library: https://github.com/letuananh/lelesk
# :copyright: (c) 2014 Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.


import logging
import os.path
import operator
from collections import defaultdict as dd
from collections import namedtuple

import nltk
from nltk.corpus import stopwords
from texttaglib.chirptext import FileHelper
from texttaglib.chirptext import Counter, Timer, uniquify, header, TextReport
from texttaglib.puchikarui import Schema

from .config import LLConfig
from .util import ptpos_to_wn, PUNCS
from yawlib import SynsetID, YLConfig
from yawlib import GWordnetSQLite as GWNSQL
from yawlib import WordnetSQL as WSQL

# -----------------------------------------------------------------------
# CONFIGURATION
# -----------------------------------------------------------------------

ScoreTup = namedtuple('Score', 'candidate score freq'.split())
WSDCandidate = namedtuple('WSDCandidate', 'id synset tokens'.split())


class LeLeskWSD:
    ''' Le's LESK algorithm for Word-Sense Disambiguation
    '''
    def __init__(self, wng_db_loc=None, wn30_loc=None, verbose=False, dbcache=None):
        logging.getLogger(__name__).debug("Initializing LeLeskWSD object ...")
        self.wng_db_loc = wng_db_loc if wng_db_loc else YLConfig.GWN30_DB
        self.wn30_loc = wn30_loc if wn30_loc else YLConfig.WNSQL30_PATH
        self.gwn = GWNSQL(self.wng_db_loc)
        self.wn = WSQL(self.wn30_loc)
        self.__stopwords = None  # this should be loaded only once
        self.PUNCS = set(PUNCS)

        if dbcache and not isinstance(dbcache, LeskCache):
            dbcache = LeskCache(dbcache)
        self.dbcache = dbcache
        self._lemmatizer = None
        self.candidates_cache = {}
        self.verbose = verbose
        self.word_cache = {}

        # database context for faster access
        self.__dbcache_ctx = None
        self.__gwn_ctx = None
        self.__wn_ctx = None
        logging.getLogger(__name__).debug("LeLeskWSD object has been initialized ...")

    def connect(self):
        ''' Use a single database connection for DB access '''
        self.disconnect()
        if self.dbcache is not None:
            self.__dbcache_ctx = self.dbcache.db.ctx()
        self.__gwn_ctx = self.gwn.ctx()
        self.__wn_ctx = self.wn.ctx()

    def disconnect(self):
        if self.__dbcache_ctx is not None:
            self.__dbcache_ctx.close()
            self.__dbcache_ctx = None
        if self.__gwn_ctx is not None:
            self.__gwn_ctx.close()
            self.__gwn_ctx = None
        if self.__wn_ctx is not None:
            self.__wn_ctx.close()
            self.__wn_ctx = None

    @property
    def stopwords(self):
        if self.__stopwords is None:
            self.__stopwords = stopwords.words('english')
            self.__stopwords.extend(PUNCS)
        return self.__stopwords

    @property
    def lemmatizer(self):
        if self._lemmatizer is None:
            self._lemmatizer = nltk.WordNetLemmatizer()
        return self._lemmatizer

    def lemmatize(self, words):
        ''' Return a list of triplets (surface, pos, lemma) '''
        tags = nltk.pos_tag(words)
        tokens = [(w, pos, self.lemmatizer.lemmatize(w, pos=ptpos_to_wn(pos, default='n'))) for w, pos in tags]
        return tokens  # [(surface, tag, lemma)]

    def lemmatize_ttl(self, sent):
        tags = nltk.pos_tag([t.text for t in sent.tokens])
        for token, (surface, pos) in zip(sent, tags):
            token.pos = pos
            token.lemma = self.lemmatizer.lemmatize(surface, pos=ptpos_to_wn(pos, default='n'))
        return sent

    def tokenize(self, sentence_text):
        return nltk.word_tokenize(sentence_text)

    def smart_synset_search(self, lemma, pos, deep_select=False):
        sses = self.gwn.search(lemma=lemma, pos=pos, deep_select=deep_select, ctx=self.__gwn_ctx)
        if len(sses) == 0:
            # try replace '-' with space
            sses = self.gwn.search(lemma=lemma.replace('-', ' '), pos=pos, ctx=self.__gwn_ctx)
        return sses

    def build_lelesk_for_word(self, a_word, pos=None, deep_select=False):
        cache_key = (a_word, pos)
        if cache_key in self.word_cache:
            return self.word_cache[cache_key]
        synsets = self.smart_synset_search(lemma=a_word, pos=pos, deep_select=deep_select)
        candidates = self.build_candidates(synsets)
        self.word_cache[cache_key] = candidates
        return candidates

    def build_candidates(self, synsets):
        candidates = []
        for idx, ss in enumerate(synsets):
            tokens = self.build_lelesk_set(ss.ID)
            candidates.append(WSDCandidate(idx + 1, ss, tokens))
        return candidates

    def build_lelesk_set(self, a_sid):
        sid_obj = SynsetID.from_string(a_sid)
        if self.dbcache is not None:
            # try to fetch from DB then ...
            lelesk_tokens = self.dbcache.select(sid_obj, ctx=self.__dbcache_ctx)
            if lelesk_tokens:
                return lelesk_tokens
        lelesk_tokens = []
        ss = self.gwn.get_synset(a_sid, ctx=self.__gwn_ctx)
        lelesk_tokens.extend(ss.get_tokens())
        lelesk_tokens.extend(ss.get_gramwords())
        # Retrieving tagged synsets
        sscol = self.gwn.get_by_keys(ss.get_tags(), ctx=self.__gwn_ctx)
        for s in sscol:
            lelesk_tokens.extend(s.get_tokens())
            lelesk_tokens.extend(s.get_gramwords())

        # Get hypehypo information from WordNet 30 DB
        ssids = self.wn.hypehypo(ss.ID, ctx=self.__wn_ctx)
        # Get synsets from Gloss WordNet
        sscol = self.gwn.get_synsets(ssids, ctx=self.__gwn_ctx)
        # dump_synsets(ss)
        for s in sscol:
            lelesk_tokens.extend(s.get_tokens())
            lelesk_tokens.extend(s.get_gramwords())

        uniquified_lelesk_tokens = [w for w in uniquify(lelesk_tokens) if w not in self.stopwords]
        # try to cache this token list ...
        if self.dbcache:
            self.dbcache.cache(sid_obj, uniquified_lelesk_tokens)
        return uniquified_lelesk_tokens

    def prepare_data(self, sentence_text, remove_stop_words=True):
        '''Given a sentence as a raw text string, perform tokenization, lemmatization
        '''
        # Tokenization
        words = self.tokenize(sentence_text)
        # Lemmatization
        tokens = self.lemmatize(words)
        if remove_stop_words:
            return [t for t in tokens if t[0] not in self.stopwords and t[2] not in self.stopwords]
        else:
            return tokens

    def lelesk_wsd(self, word, sentence_text='', expected_sense='', lemmatizing=True, pos=None, context=None, synsets=None, remove_stop_words=True, **kwargs):
        ''' Perform Word-sense disambiguation with extended simplified LESK and annotated WordNet 3.0
        '''
        # 1. Retrieve candidates for the given word
        if not synsets:
            if (word, pos) not in self.candidates_cache:
                self.candidates_cache[(word, pos)] = self.build_lelesk_for_word(word, pos=pos)
            candidates = self.candidates_cache[(word, pos)]
        else:
            candidates = self.build_candidates(synsets)

        # 2. Calculate overlap between the context and each given word
        if not context:
            context = uniquify([x[2] for x in self.prepare_data(sentence_text, remove_stop_words=remove_stop_words)])
        else:
            if remove_stop_words:
                context_set = set(x for x in context if self.stopwords and x not in self.stopwords)
            else:
                context_set = set(context)
            
        scores = []

        logging.getLogger(__name__).info("candidate for {} => {}".format(word, list(c.synset for c in candidates)))
        for candidate in candidates:
            candidate_set = set(candidate.tokens)
            score = len(context_set.intersection(candidate_set))
            scores.append(ScoreTup(candidate, score, self.wn.get_tagcount(candidate.synset.ID.to_wnsql(), ctx=self.__wn_ctx)))
            # scores.append([candidate, score, candidate.sense.tagcount])
        scores.sort(key=operator.itemgetter(1, 2))
        scores.reverse()
        return scores

    def mfs_wsd(self, word, sentence_text, expected_sense='', lemmatizing=True, pos=None, synsets=None, **kwargs):
        '''Perform Word-sense disambiguation with just most-frequent senses
        '''
        # 1. Retrieve candidates for the given word
        if not synsets:
            synsets = self.smart_synset_search(lemma=word, pos=pos)
        candidates = []
        for idx, ss in enumerate(synsets):
            candidates.append(WSDCandidate(idx + 1, ss, []))

        scores = []
        for candidate in candidates:
            freq = self.wn.get_tagcount(candidate.synset.ID.to_wnsql(), ctx=self.__wn_ctx)
            score = freq
            scores.append(ScoreTup(candidate, score, freq))
        scores.sort(key=operator.itemgetter(1))
        scores.reverse()
        return scores


class LeskCacheSchema(Schema):
    def __init__(self, data_source=None, setup_file=LLConfig.LELESK_CACHE_DB_INIT_SCRIPT):
        Schema.__init__(self, data_source, setup_file=setup_file)
        # tokens: synsetid token
        self.add_table('tokens', 'synsetid token'.split())
        # synset: synsetid offset pos synsetid_wn30 freq
        self.add_table('synset', 'synsetid offset pos synsetid_gwn freq'.split())
        # sensekey: synsetid sensekey
        self.add_table('sensekey', 'synsetid sensekey'.split())
        # term: synsetid term
        self.add_table('term', 'synsetid term'.split())


class LeskCache:
    def __init__(self, db_file=None, wsd=None, debug_dir=None):
        ''' Create an instance of LeskCache

        Arguments:
            wsd       -- An instance of LeLeskWSD
            db_file   -- Path to Cache DB file (default to ./data/lesk_cache.db)
            debug_dir -- Details of how tokens are generated for synset will be stored here.
            (Default values are defined in config.py)
        '''
        self.wsd = wsd
        self.db_file = db_file if db_file else LLConfig.LELESK_CACHE_DB_LOC
        self.debug_dir = debug_dir if debug_dir else LLConfig.LELESK_CACHE_DEBUG_DIR
        self.db = LeskCacheSchema(self.db_file)
        # try to setup DB if needed
        if self.db_file is not None:
            # Create dir if needed
            logging.getLogger(__name__).info("LeskCache DB is located at {}".format(self.db_file))
            if self.db_file != ':memory:':
                FileHelper.create_dir(os.path.dirname(self.db_file))

    def info(self):
        header("Pre-generate LESK tokens for all synsets for faster WSD")
        print("Path to WordNet Gloss Corpus: %s" % self.wsd.wng_db_loc)
        print("Path to WordNet 3.0 SQLite  : %s" % self.wsd.wn30_loc)
        print("Path to LeLesk cache DB     : %s" % self.db_file)
        print("Debug info will be stored in: %s" % self.debug_dir)
        print("--")
        print("")

    def cache(self, synsetid, tokens):
        # delete tokens first
        self.db.ds.execute('DELETE FROM tokens WHERE synsetid=?', [str(synsetid)])
        with self.db.ctx() as ctx:
            for token in tokens:
                self.db.tokens.insert(str(synsetid), token, ctx=ctx)

    def select(self, synsetid, ctx=None):
        result = self.db.tokens.select('synsetid=?', (str(synsetid),), ctx=ctx)
        if result:
            return [x.token for x in result]
        else:
            return None

    def validate(self):
        gwn = self.wsd.gwn
        wn = self.wsd.wn
        t = Timer()

        print("Validating WordNet databases ...")
        # Comparing synsets
        wn_synsets = wn.get_all_synsets()
        synsetids = set()
        for row in wn_synsets:  # 'synsetid', 'lemma', 'sensekey', 'tagcount'
            synsetids.add(row.synsetid)
        print("Synsets in WordNet SQLite: %s" % len(synsetids))

        t.start("Selecting from WordNet Gloss Corpus")
        gwn_synsets = gwn.all_synsets(deep_select=False)
        offset_map = {}
        for synset in gwn_synsets:
            offset_map[synset.offset] = synset.id
        t.end("Done GWN")
        print("Synsets in WordNet Gloss Corpus: %s" % len(gwn_synsets))

        c = Counter()
        for synsetid in synsetids:
            if str(synsetid)[1:] in offset_map:
                c.count("Match")
            else:
                c.count("Nah")
        c.summarise()

        if c['Match'] != len(gwn_synsets):
            print("Invalid WordNet version detected. Execution aborted!")
            return
        else:
            print("GWN DB and WN30 SQLite DB matched. Proceed with data generation")
        print('')

        t.start('Caching WN30 sensekey map')
        wnsks = wn.get_all_sensekeys()
        wn_skmap = {}
        wn_sidmap = dd(list)
        # map by sensekeys and synsetid
        for item in wnsks:
            wn_skmap[item.sensekey] = item.synsetid
            wn_sidmap[str(item.synsetid)[1:]].append(item.sensekey)
        t.end("Done WN30")

        t.start('Caching GWN sensekey map')
        gwn_ss = gwn.get_all_sensekeys()
        gwn_skmap = {}
        for item in gwn_ss:
            gwn_skmap[item.sensekey] = item.ID
        t.end("Done GWN")

        t.start('Caching GWN tagged sensekey')
        gwn_tags = gwn.get_all_sensekeys_tagged()
        t.end("Done tagged sensekey")

        print("wn30 sensekeys: %s" % len(wnsks))
        print("gwn synsets   : %s" % len(gwn_ss))
        print("All tagged sk : %s" % len(gwn_tags))

        c = Counter()
        for tag in gwn_tags:
            if tag not in gwn_skmap:
                print("sk [%s] does not exist in GWN" % tag)
                c.count("GWN Not Found")
            else:
                c.count("GWN Found")
            if tag not in wn_skmap:
                if tag in gwn_skmap:
                    gwn_sid = gwn_skmap[tag][1:]
                    # print("Searching %s" % (gwn_sid))
                    if gwn_sid in wn_sidmap:
                        candidates = wn_sidmap[gwn_sid]
                        derivative = None
                        for cand in candidates:
                            if tag.split('%')[0] == cand.split('%')[0]:
                                derivative = cand
                                break
                        if derivative:
                            c.count("WN30 Found derivative")
                        else:
                            # not found at all
                            c.count("WN30 Not Found At all")
                            print("sk [%s] does not exist in WN30 at all ..." % tag)
                else:
                    c.count("WN30 & GWN Not Found")
                    print("sk [%s] does not exist in WN30" % tag)
            else:
                c.count("WN30 Found")
        c.summarise()

    def generate(self):
        synsets = self.wsd.gwn.all_synsets(deep_select=False)
        with self.db.ds.open() as ctx:
            total_synsets = len(synsets)
            for idx, synset in enumerate(synsets):
                logging.getLogger(__name__).info("Generating tokens for %s (%s/%s)" % (synset.id, idx, total_synsets))
                debug_file = TextReport(os.path.join(self.debug_dir, synset.offset + '.txt'))
                tokens = self.wsd.build_lelesk_set(synset.id, debug_file)
                for token in tokens:
                    self.db.tokens.insert(synset.id, token, ctx=ctx)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
LeLESK configuration module
'''

# This code is a part of lelesk library: https://github.com/letuananh/lelesk
# :copyright: (c) 2014 Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

import os


class LLConfig:
    LELESK_DIR = os.path.dirname(os.path.abspath(__file__))
    LELESK_CACHE_DB_INIT_SCRIPT = os.path.join(LELESK_DIR, 'script', 'lesk_cache.sql')
    LELESK_CACHE_DB_LOC = os.path.expanduser('./data/temp/lesk_cache.db')
    LELESK_CACHE_DEBUG_DIR = os.path.expanduser('./data/temp/debug')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Word Sense Disambiguation Toolkit

Usage:
    # Build lelesk set
    python3 wsdtk.py -g ~/wordnet/gwn.db -w ~/wordnet/sqlite-30.db -l v01775535

    # Test retrieving sense candidates for a word
    python3 wsdtk.py --candidates "love" --pos "v"

    # test WSD batch mode
    python3 wsdtk.py -b data/test.txt -o data/test_report.txt -r data/test_debug.txt
"""

# This code is a part of lelesk library: https://github.com/letuananh/lelesk
# :copyright: (c) 2014 Le Tuan Anh <tuananh.ke@gmail.com>
# :license: MIT, see LICENSE for more details.

import os.path
from collections import namedtuple

from texttaglib.chirptext.leutile import Timer, Counter, Table
from texttaglib.chirptext.cli import CLIApp, setup_logging
from texttaglib.chirptext import chio
from texttaglib.chirptext import ttl
from yawlib import YLConfig

from .main import LeLeskWSD, LeskCache
from .util import ptpos_to_wn

# -----------------------------------------------------------------------

OutputLine = namedtuple('OutputLine', 'results word correct_sense suggested_sense sentence_text'.split())
setup_logging('logging.json', 'logs')


def generate_tokens(wsd):
    ''' Pre-generate LESK tokens for all synsets for faster WSD
    '''
    lesk_cache = LeskCache(wsd)
    lesk_cache.info()
    lesk_cache.generate()
    print("Done!")


# -----------------------------------------------------------------------

def batch_wsd(infile_loc, wsd_obj, outfile_loc=None, method='lelesk', use_pos=False, assume_perfect_POS=False, lemmatizing=False, pretokenized=False):
    ''' Perform WSD in batch mode (input is a text file with each line is a sentence)
        Arguments:
            infile_loc         -- path to input file (Tab separated file)
            wsd_obj            -- WSD component (e.g. LeLeskWSD object)
            outfile_loc        -- Path to output (log) file
            method             -- 'lelesk' or 'mfs'
            use_pos            -- Use part-of-speech or not
            assume_perfect_POS -- If True then use POS from input file
            lemmatizing        -- Lemmatize tokens
            pretokenized       -- Use context's tokens from input file
    '''
    tbatch = Timer()  # total time used to process this batch
    tbatch.start("Batch WSD started | Method=%s | File = %s" % (method, infile_loc))
    t = Timer()
    t.start("Reading file %s" % infile_loc)
    lines = open(os.path.expanduser(infile_loc)).readlines()
    t.end("File has been loaded")

    c = Counter('Match InTop3 Wrong NoSense TotalSense'.split())
    outputlines = []
    # Counters for different type of words
    match_count = Counter()
    top3_count = Counter()
    wrong_count = Counter()
    nosense_count = Counter()

    # sample line in input file:
    # adventure 00796315-n  n   The Adventure of the Speckled Band  the|adventure|of|the|speckle|band
    total_lines = len(lines)
    processed = 0
    for line in lines:
        processed += 1
        if line.startswith('#'):
            # it is a comment line
            continue
        parts = line.split('\t')
        if parts:
            context = None
            if len(parts) == 3: 
                word = parts[0].strip()
                correct_sense = parts[1].strip()
                pos = None
                sentence_text = parts[2].strip()
            elif len(parts) == 4:
                word = parts[0].strip()
                correct_sense = parts[1].strip()
                pos = parts[2].strip()
                sentence_text = parts[3].strip()
            elif len(parts) == 5:
                word = parts[0].strip()
                correct_sense = parts[1].strip()
                pos = parts[2].strip()
                sentence_text = parts[3].strip()
                context = parts[4].strip().split('|')
            if pos in ('', 'x'):
                pos = None

            if assume_perfect_POS:
                pos = correct_sense[-1]
            if not use_pos:
                # if use choose to ignore POS
                pos = None

            results = 'X'
            # Perform WSD on given word & sentence
            t.start('Analysing word ["%s"] on sentence: %s' % (word, sentence_text))
            if method == 'mfs':
                scores = wsd_obj.mfs_wsd(word, sentence_text, correct_sense, lemmatizing=lemmatizing, pos=pos)
            else:
                if pretokenized and context and len(context) > 0:
                    scores = wsd_obj.lelesk_wsd(word, sentence_text, correct_sense, lemmatizing=lemmatizing, pos=pos, context=context)
                else:
                    scores = wsd_obj.lelesk_wsd(word, sentence_text, correct_sense, lemmatizing=lemmatizing, pos=pos)
            suggested_senses = [score.candidate.synset.ID for score in scores[:3]]
            # c.count("TotalSense")

            if correct_sense in YLConfig.NTUMC_PRONOUNS:
                c.count("IGNORED")
                continue
            else:
                c.count("TotalSense")

            if len(suggested_senses) == 0:
                nosense_count.count(correct_sense + '\t' + word)
                c.count("NoSense")
                results = '_'
            elif correct_sense == suggested_senses[0]:
                match_count.count(correct_sense + '\t' + word)
                c.count("Match")
                results = 'O'
            elif correct_sense in suggested_senses:
                top3_count.count(correct_sense + '\t' + word)
                c.count("InTop3")
                results = 'V'
            else:
                wrong_count.count(correct_sense + '\t' + word)
                c.count("Wrong")

            # write to output file
            if len(suggested_senses) > 0:
                outputlines.append(OutputLine(results, word, correct_sense, suggested_senses[0], sentence_text))
            else:
                outputlines.append(OutputLine(results, word, correct_sense, "", sentence_text))
            t.end('Analysed word ["%s"] (%s/%s completed)' % (word, processed, total_lines))
    print("")
    tbatch.start("Batch WSD finished | Method=%s | File = %s" % (method, infile_loc))

    if outfile_loc:
        print("Writing output file ==> %s..." % (outfile_loc,))

        with open(outfile_loc, 'w') as outfile:
            outfile.write("Sections\n")
            outfile.write("::SENTENCES::\n")
            outfile.write("::MATCH-TOKENS::\n")
            outfile.write("::TOP3-TOKENS::\n")
            outfile.write("::WRONG-TOKENS::\n")
            outfile.write("::NOSENSE-TOKENS::\n")
            outfile.write("::SUMMARY::\n")
            outfile.write(("-" * 20) + '\n')

            outfile.write("::SENTENCES::\n")
            tbl = Table()
            tbl.add_row("results word correct_sense suggested_sense sentence_text".split())
            for line in outputlines:
                tbl.add_row((line.results ,line.word ,line.correct_sense ,line.suggested_sense ,line.sentence_text))
            print_tbl = lambda x: outfile.write('%s\n' % x)
            # write table
            tbl.print(print_func=print_tbl)
            
            outfile.write("\n")
            # dump counter tokens
            dump_counter(match_count, outfile, '::MATCH-TOKENS::')
            dump_counter(top3_count, outfile, '::TOP3-TOKENS::')
            dump_counter(wrong_count, outfile, '::WRONG-TOKENS::')
            dump_counter(nosense_count, outfile, '::NOSENSE-TOKENS::')
            # write summary
            totalcount = len(match_count) + len(top3_count) + len(wrong_count) + len(nosense_count)
            outfile.write("\n")
            outfile.write("::SUMMARY::\n")
            outfile.write("%s\n" % (tbatch))
            outfile.write("\n")
            outfile.write("| Information                         |    Instance | Classes |\n")
            outfile.write("|:------------------------------------|--------:|-----------:\n")
            outfile.write("| Correct sense ranked the first      |   %s | %s |\n" % (str(c['Match']).rjust(5, ' '), str(len(match_count)).rjust(5, ' ')))
            outfile.write("| Correct sense ranked the 2nd or 3rd |   %s | %s |\n" % (str(c['InTop3']).rjust(5, ' '), str(len(top3_count)).rjust(5, ' ')))
            outfile.write("| Wrong                               |   %s | %s |\n" % (str(c['Wrong']).rjust(5, ' '), str(len(wrong_count)).rjust(5, ' ')))
            outfile.write("| NoSense                             |   %s | %s |\n" % (str(c['NoSense']).rjust(5, ' '), str(len(nosense_count)).rjust(5, ' ')))
            outfile.write("| TotalSense                          |   %s | %s |\n" % (str(c['TotalSense']).rjust(5, ' '), str('').rjust(5, ' '))) # totalcount is wrong!
    print("Batch job finished")


def dump_counter(counter, file_obj, header):
    tbl = Table()
    tbl.add_row(["Synset ID", "Lemma", "Count"])
    items = counter.sorted_by_count()
    for k, v in items:
        tbl.add_row(k.split('\t') + [v])

    print_tbl = lambda x: file_obj.write('%s\n' % x)
    file_obj.write("\n")

    # write header
    file_obj.write("\n%s\n" % header)
    # write table
    tbl.print(print_func=print_tbl)


def get_lelesk_set(cli, args):
    ''' Get hyponyms and hypernyms given a synsetID '''
    wsd = build_wsd_object(cli, args)
    print(wsd.build_lelesk_set(args.synsetid))


def build_wsd_object(cli, args):
    wsd = LeLeskWSD(args.glosswn, args.wnsql, verbose=not args.quiet, dbcache=LeskCache())
    return wsd


def tokenize_text(cli, args):
    wsd = build_wsd_object(cli, args)
    tokens = wsd.prepare_data(args.text)
    for surface, pos, lemma in tokens:
        print(surface, pos, lemma)


def wsd_text(cli, args):
    ''' Perform WSD on a given WORD in a given CONTEXT'''
    sent = wsd_sent(ttl.Sentence(text=args.context), cli, args)
    print("Text: {}".format(sent.text))
    if args.debug:
        print("Tokens: {}".format(', '.join('{}/{}/{}'.format(x, x.pos, x.lemma) for x in sent.tokens)))
    else:
        print("Tokens: {}".format(', '.join(str(x) for x in sent.tokens)))
    for c in sent.concepts:
        print(c)


def wsd_document(doc, cli, args):
    """ Batch perform WSD """
    wsd = build_wsd_object(cli, args)
    if args.ttl_format == ttl.MODE_JSON:
        _writer = ttl.JSONWriter.from_path(args.output)
    else:
        _writer = ttl.TxtWriter.from_path(args.output)
    stopwords = set(wsd.stopwords)
    if not args.method or args.method.lower() in ('lesk', 'lelesk'):
        wsd_method = "LELESK"
        wsd_func = wsd.lelesk_wsd
    elif args.method.lower() == 'mfs':
        wsd_method = "MFS"
        wsd_func = wsd.mfs_wsd
    else:
        print("Unknown WSD method: {}".format(args.method))
        exit()
    t = Timer()
    t.start("WSD method: {}".format(wsd_method))
    for idx, sent in enumerate(doc):
        if args.topk and args.topk - 1 < idx:
            break
        print("Sent {}/{}: {} ".format(idx + 1, len(doc), sent.text))
        wsd_sent(sent, cli, args, wsd, stopwords, wsd_method, wsd_func)
        # write sentence
        _writer.write_sent(sent)
    print("Output was written to {}".format(args.output))
    t.stop("Done WSD")


def wsd_file(cli, args):
    """ WSD a text file """
    if args.format == "ttl":
        return wsd_ttl(cli, args)
    elif args.format == "txt":
        if os.path.isfile(args.input):
            lines = chio.read_file(args.input).splitlines()
            doc = ttl.Document()
            for line in lines:
                doc.new_sent(line)
            wsd_document(doc, cli, args)
        else:
            print(f"Error. Input file not found! ({args.input})")


def wsd_ttl(cli, args):
    ''' Perform WSD on a TTL document '''
    doc = ttl.read(args.input, mode=args.ttl_format)  # input TTL
    wsd_document(doc, cli, args)


def wsd_candidates(sent, cli, args, wsd=None, stopwords=None, remove_stop_words=True, **kwargs):
    if wsd is None:
        wsd = build_wsd_object(cli, args)
    if stopwords is None:
        stopwords = set(wsd.stopwords)
    if not sent.tokens:
        sent.tokens = wsd.tokenize(sent.text)
    # lemmatize if needed
    if not args.nolemmatize:
        wsd.lemmatize_ttl(sent)
    # build WSD context
    context = set(token.text.lower() for token in sent.tokens)
    context.update(token.lemma.lower() for token in sent.tokens if token.lemma)
    if remove_stop_words:
        context = context - stopwords
    cli.logger.debug("Sent #{} tokens: {}".format(sent.ID, [(t.text, t.lemma, t.pos) for t in sent]))
    cli.logger.debug("Sent #{} context: {}".format(sent.ID, context))
    for token in sent:
        if token.lemma in wsd.PUNCS or token.text in wsd.PUNCS:
            continue
        elif token.pos and token.pos in ('PRP', '.'):
            continue
        # no need to remove stopwords again, they are removed above
        # output = wsd_func(token.lemma if token.lemma else token.text, sent.text, pos=ptpos_to_wn(token.pos), context=context, remove_stop_words=False)
        synsets = wsd.smart_synset_search(lemma=token.lemma if token.lemma else token.text, pos=ptpos_to_wn(token.pos))
        if synsets:
            for ss in synsets:
                sent.new_concept(ss.ID, clemma=token.lemma if token.lemma else token.text, tokens=[token])
                if not args.notag:
                    # also make tags
                    sent.new_tag(str(ss.ID), token.cfrom, token.cto, tagtype='WN')
            cli.logger.debug("  {}/tk:[{}] => {}".format('#{}'.format(sent.ID) or '', (token.text, token.lemma, token.pos), list(synsets)))
    return sent


def wsd_sent(sent, cli, args, wsd=None, stopwords=None, wsd_method=None, wsd_func=None, remove_stop_words=True):
    if wsd is None:
        wsd = build_wsd_object(cli, args)
    if wsd_method is None or wsd_func is None:
        if not args.method or args.method.lower() in ('lesk', 'lelesk'):
            wsd_method = "LELESK"
            wsd_func = wsd.lelesk_wsd
        elif args.method.lower() == 'mfs':
            wsd_method = "MFS"
            wsd_func = wsd.mfs_wsd
    if stopwords is None:
        stopwords = set(wsd.stopwords)
    if not sent.tokens:
        sent.tokens = wsd.tokenize(sent.text)
    # lemmatize if needed
    if not args.nolemmatize:
        wsd.lemmatize_ttl(sent)
    # build WSD context
    context = set(token.text.lower() for token in sent.tokens)
    context.update(token.lemma.lower() for token in sent.tokens if token.lemma)
    if remove_stop_words:
        context = context - stopwords
    cli.logger.debug("Sent #{} tokens: {}".format(sent.ID, [(t.text, t.lemma, t.pos) for t in sent]))
    cli.logger.debug("Sent #{} context: {}".format(sent.ID, context))
    for token in sent:
        if token.lemma in wsd.PUNCS or token.text in wsd.PUNCS:
            continue
        elif token.pos and token.pos in ('PRP', '.'):
            continue
        # no need to remove stopwords again, they are removed above
        output = wsd_func(token.lemma if token.lemma else token.text, sent.text, pos=ptpos_to_wn(token.pos), context=context, remove_stop_words=False)
        if output:
            c = output[0].candidate
            concept = sent.new_concept(c.synset.ID, clemma=token.lemma if token.lemma else token.text, tokens=[token])
            if not args.notag:
                # also make tags
                sent.new_tag(str(c.synset.ID), token.cfrom, token.cto, tagtype='WN')
            cli.logger.debug("  {}/{}/ tk:[{}] => {}".format('#{}'.format(sent.ID) or '', wsd_method, (token.text, token.lemma, token.pos), concept))
    return sent


def find_candidates(cli, args):
    ''' Retrieve Word Sense Disambiguation candidates for a word '''
    wsd = build_wsd_object(cli, args)
    candidates = wsd.build_lelesk_for_word(args.word, args.pos, deep_select=True)
    for candidate in candidates:
        print(candidate.id, candidate.synset.ID, candidate.synset.lemmas, candidate.synset.definition)


# -----------------------------------------------------------------------

def main():
    '''Main entry of WSD toolkit
    '''
    app = CLIApp(desc='LeLesk - Word-Sense Disambiguation Toolkit', logger=__name__)
    # Positional argument(s)
    app.parser.add_argument('-w', '--wnsql', help='Location to WordNet 3.0 SQLite database', default=YLConfig.WNSQL30_PATH)
    app.parser.add_argument('-g', '--glosswn', help='Location to Gloss WordNet SQLite database', default=YLConfig.GWN30_DB)

    task = app.add_task('lelesk', func=get_lelesk_set)
    task.add_argument('synsetid', help='Synset ID')

    task = app.add_task('candidates', func=find_candidates)
    task.add_argument('word', help='Word to search for synsets')
    task.add_argument('--pos', default=None)
    task.add_argument('--show_tokens', action="store_true")

    task = app.add_task('tokenize', func=tokenize_text)
    task.add_argument('text', help='Sentence text to analyse')

    task = app.add_task('wsd', func=wsd_text)
    task.add_argument('context', help='Context to perform WSD')
    task.add_argument('--word', help='word to perform WSD')
    task.add_argument('--notag', help='Also use sentence level tags for annotations', action='store_true')
    task.add_argument('--nolemmatize', help='Do not perform lemmatization', action='store_true')
    task.add_argument('-m', '--method', help='WSD method (mfs/lelesk)', choices=['mfs', 'lelesk'])
    task.add_argument('--debug', action='store_true')

    task = app.add_task('file', func=wsd_file)
    task.add_argument('input', help='Input file')
    task.add_argument('output', help='Output file (always in TTL format)')
    task.add_argument('-n', '--topk', help='Only process top k sentences', type=int)
    task.add_argument('--notag', help='Also use sentence level tags for annotations', action='store_true')
    task.add_argument('--nolemmatize', help='Do not perform lemmatization', action='store_true')
    task.add_argument('-m', '--method', help='WSD method (mfs/lelesk)', choices=['mfs', 'lelesk', 'lesk'], default='lelesk')
    task.add_argument('-f', '--format', help="File format (TTL, txt)", choices=["txt", "ttl"], default='txt')
    task.add_argument('--ttl_format', help='TTL format', default=ttl.MODE_TSV, choices=[ttl.MODE_JSON, ttl.MODE_TSV])
    
    task = app.add_task('ttl', func=wsd_ttl)
    task.add_argument('input', help='TTL profile')
    task.add_argument('output', help='Output TTL profile')
    task.add_argument('-n', '--topk', help='Only process top k sentences', type=int)
    task.add_argument('--notag', help='Also use sentence level tags for annotations', action='store_true')
    task.add_argument('--nolemmatize', help='Do not perform lemmatization', action='store_true')
    task.add_argument('-m', '--method', help='WSD method (mfs/lelesk)', choices=['mfs', 'lelesk', 'lesk'], default='lelesk')
    task.add_argument('--ttl_format', help='TTL format', default=ttl.MODE_TSV, choices=[ttl.MODE_JSON, ttl.MODE_TSV])

    task = app.add_task('cand', func=wsd_candidates)
    task.add_argument('input', help='TTL profile')
    task.add_argument('output', help='Output TTL profile')
    task.add_argument('-n', '--topk', help='Only process top k sentences', type=int)
    task.add_argument('--notag', help='Also use sentence level tags for annotations', action='store_true')
    task.add_argument('--nolemmatize', help='Do not perform lemmatization', action='store_true')
    task.add_argument('-m', '--method', help='WSD method (mfs/lelesk)', choices=['mfs', 'lelesk', 'lesk'], default='lelesk')
    task.add_argument('--ttl_format', help='TTL format', default=ttl.MODE_TSV, choices=[ttl.MODE_JSON, ttl.MODE_TSV])
    app.run()


if __name__ == "__main__":
    main()
